@@ -1,4 +1,5 @@
 from redis.asyncio.client import Redis, StrictRedis
+from redis.asyncio.cluster import RedisCluster
 from redis.asyncio.connection import (
     BlockingConnectionPool,
     Connection,
@@ -6,6 +7,7 @@ from redis.asyncio.connection import (
     SSLConnection,
     UnixDomainSocketConnection,
 )
+from redis.asyncio.parser import CommandsParser
 from redis.asyncio.sentinel import (
     Sentinel,
     SentinelConnectionPool,
@@ -13,6 +15,7 @@ from redis.asyncio.sentinel import (
     SentinelManagedSSLConnection,
 )
 from redis.asyncio.utils import from_url
+from redis.backoff import default_backoff
 from redis.exceptions import (
     AuthenticationError,
     AuthenticationWrongNumberOfArgsError,
@@ -35,15 +38,18 @@ __all__ = [
     "BlockingConnectionPool",
     "BusyLoadingError",
     "ChildDeadlockedError",
+    "CommandsParser",
     "Connection",
     "ConnectionError",
     "ConnectionPool",
     "DataError",
     "from_url",
+    "default_backoff",
     "InvalidResponse",
     "PubSubError",
     "ReadOnlyError",
     "Redis",
+    "RedisCluster",
     "RedisError",
     "ResponseError",
     "Sentinel",
